////////////////////////////////////////////////////////////////
// File - P9054_LIB.C
//
// Library for 'WinDriver for PLX 9054' API. 
// The basic idea is to get a handle for the board
// with P9054_Open() and use it in the rest of the program
// when calling WD functions.  Call P9054_Close() when done.
// 
// Copyright (c) 2003 - 2004 Jungo Ltd.  http://www.jungo.com
// 
////////////////////////////////////////////////////////////////

#include "p9054_lib.h"
#include "/WinDriver/include/windrvr_int_thread.h"
#include "/WinDriver/include/status_strings.h"
#include "/WinDriver/include/utils.h"
#if !defined(__KERNEL__) 
    #include <stdio.h>
#endif

// this string is set to an error message, if one occurs
CHAR P9054_ErrorString[1024];

// internal data structures and enum definitions 
enum { P9054_DMA_CHANNEL_SHIFT = 0x14 }; // shift in address between channels 0 and 1 of DMA
enum { P9054_EEPROM_DELAY = 20000, P9054_SOFT_RESET_DELAY = 5000000 };

typedef struct P9054_DMA_STRUCT{
    WD_DMA dmaList;
    P9054_DMA_CHANNEL dmaChannel;
    WD_DMA dma;
} P9054_DMA_STRUCT;

enum { P9054_MODE_DESC       = (int)0xF9000140 };
enum { P9054_MODE_DESC_BYTE  = 0x00000000 };
enum { P9054_MODE_DESC_WORD  = 0x00010001 };
enum { P9054_MODE_DESC_DWORD = 0x00030003 };

typedef struct
{
    WD_INTERRUPT Int;
    HANDLE hThread;
    WD_TRANSFER Trans[2];
    P9054_INT_HANDLER funcIntHandler;
} P9054_INTERRUPT;

typedef struct 
{
    DWORD dwLocalBase;
    DWORD dwMask;
    DWORD dwBytes;
    KPTR dwAddr;
    DWORD dwAddrDirect;
    BOOL  fIsMemory;
} P9054_ADDR_DESC;

typedef struct P9054_STRUCT
{
    HANDLE hWD;
    WD_CARD cardLock;
    WD_PCI_SLOT pciSlot;
    WD_CARD_REGISTER cardReg;
    P9054_ADDR_DESC addrDesc[AD_PCI_BARS];
    P9054_ADDR addrSpace;
    P9054_INTERRUPT Int;
} P9054_STRUCT;

// internal function used by P9054_Open()
BOOL P9054_DetectCardElements(P9054_HANDLE hPlx);
// internal function used by P9054_Read... and P9054_Write... functions
void P9054_SetMode (P9054_HANDLE hPlx, DWORD dwLocalAddr);

BOOL P9054_CheckVersion(void)
{
    HANDLE hWD = INVALID_HANDLE_VALUE;
    DWORD dwStatus;
    WD_VERSION ver;

    hWD = WD_Open();
    // check if handle valid & version OK
    if (hWD==INVALID_HANDLE_VALUE) 
    {
        sprintf(P9054_ErrorString, "Failed opening " WD_PROD_NAME " device\n");
        return FALSE;
    }
    BZERO(ver);
    dwStatus = WD_Version(hWD, &ver);
    WD_Close (hWD);
    if (dwStatus)
    {
        sprintf(P9054_ErrorString, "WD_Version() failed with status 0x%lx - %s\n",
            dwStatus, Stat2Str(dwStatus));
        return FALSE;
    }
    if (ver.dwVer<WD_VER)
    {
        sprintf(P9054_ErrorString, "Incorrect " WD_PROD_NAME " version. "
            "expected %d.%02d, got %ld.%02ld\n",
            WD_VER/100, WD_VER%100, ver.dwVer/100, ver.dwVer%100);
        return FALSE;
    }
    return TRUE;
}

DWORD P9054_CountCards (DWORD dwVendorID, DWORD dwDeviceID)
{
    WD_PCI_SCAN_CARDS pciScan;
    HANDLE hWD = INVALID_HANDLE_VALUE;
    DWORD dwStatus;

    P9054_ErrorString[0] = '\0';
    if (!P9054_CheckVersion())
        return 0;
    hWD = WD_Open();
    // check if handle valid & version OK
    if (hWD==INVALID_HANDLE_VALUE) 
    {
        sprintf(P9054_ErrorString, "Failed opening " WD_PROD_NAME " device\n");
        return 0;
    }

    BZERO(pciScan);
    pciScan.searchId.dwVendorId = dwVendorID;
    pciScan.searchId.dwDeviceId = dwDeviceID;
    dwStatus = WD_PciScanCards(hWD, &pciScan);
    WD_Close(hWD);
    if (dwStatus)
    {
        sprintf(P9054_ErrorString, "WD_PciScanCards() failed with status 0x%lx - %s\n",
            dwStatus, Stat2Str(dwStatus));
        pciScan.dwCards = 0;
    }
    else if (pciScan.dwCards==0)
        sprintf(P9054_ErrorString, "no cards found\n");
    return pciScan.dwCards;
}

BOOL P9054_Open(P9054_HANDLE *phPlx, DWORD dwVendorID, DWORD dwDeviceID, DWORD nCardNum)
{
    P9054_HANDLE hPlx = (P9054_HANDLE) malloc (sizeof (P9054_STRUCT));
    DWORD dwIntStatus;
    DWORD dwStatus;

    WD_PCI_SCAN_CARDS pciScan;
    WD_PCI_CARD_INFO pciCardInfo;

    *phPlx = NULL;
    P9054_ErrorString[0] = '\0';
    if (!P9054_CheckVersion())
        return FALSE;
    BZERO(*hPlx);

    hPlx->hWD = WD_Open();

    // check if handle valid & version OK
    if (hPlx->hWD==INVALID_HANDLE_VALUE) 
    {
        sprintf(P9054_ErrorString, "Failed opening " WD_PROD_NAME " device\n");
        goto Exit;
    }

    BZERO(pciScan);
    pciScan.searchId.dwVendorId = dwVendorID;
    pciScan.searchId.dwDeviceId = dwDeviceID;
    dwStatus = WD_PciScanCards (hPlx->hWD, &pciScan);
    if (dwStatus)
    {
        sprintf(P9054_ErrorString, "WD_PciScanCards() failed with status 0x%lx - %s\n",
            dwStatus, Stat2Str(dwStatus));
        goto Exit;
    }
    if (pciScan.dwCards==0) // Found at least one card
    {
        sprintf(P9054_ErrorString, "Could not find PCI card\n");
        goto Exit;
    }
    if (pciScan.dwCards<=nCardNum)
    {
        sprintf(P9054_ErrorString, "Card out of range of available cards\n");
        goto Exit;
    }

    BZERO(pciCardInfo);
    pciCardInfo.pciSlot = pciScan.cardSlot[nCardNum];
    dwStatus = WD_PciGetCardInfo(hPlx->hWD, &pciCardInfo);
    if (dwStatus)
    {
        sprintf(P9054_ErrorString, "WD_PciGetCardInfo() failed with status 0x%lx - %s\n",
            dwStatus, Stat2Str(dwStatus));
        goto Exit;
    }
    hPlx->pciSlot = pciCardInfo.pciSlot;
    hPlx->cardReg.Card = pciCardInfo.Card;
    hPlx->cardReg.fCheckLockOnly = FALSE;
    dwStatus = WD_CardRegister(hPlx->hWD, &hPlx->cardReg);
    if (dwStatus)
    {
        sprintf(P9054_ErrorString, "WD_CardRegister() failed with status 0x%lx - %s\n",
            dwStatus, Stat2Str(dwStatus));
        goto Exit;
    }
    if (!P9054_DetectCardElements(hPlx))
    {
        sprintf(P9054_ErrorString, "Card does not have all items expected for PLX 9054\n");
        goto Exit;
    }

    // this enables target abort so it wont get stuck 
    dwIntStatus = P9054_ReadReg (hPlx, P9054_INTCSR);
    P9054_WriteReg(hPlx, P9054_INTCSR, dwIntStatus | BIT12);

    // Open finished OK
    *phPlx = hPlx;
    return TRUE;

Exit:
    // Error during Open
    if (hPlx->cardReg.hCard) 
        WD_CardUnregister(hPlx->hWD, &hPlx->cardReg);
    if (hPlx->hWD!=INVALID_HANDLE_VALUE)
        WD_Close(hPlx->hWD);
    free(hPlx);
    return FALSE;
}

void P9054_GetPciSlot(P9054_HANDLE hPlx, WD_PCI_SLOT *pPciSlot)
{
    *pPciSlot = hPlx->pciSlot;
}

UINT32 P9054_ReadPCIReg(P9054_HANDLE hPlx, DWORD dwReg)
{
    WD_PCI_CONFIG_DUMP pciCnf;
    UINT32 dwVal;

    BZERO (pciCnf);
    pciCnf.pciSlot = hPlx->pciSlot;
    pciCnf.pBuffer = &dwVal;
    pciCnf.dwOffset = dwReg;
    pciCnf.dwBytes = 4;
    pciCnf.fIsRead = TRUE;
    WD_PciConfigDump(hPlx->hWD,&pciCnf);
    return dwVal;
}

void P9054_WritePCIReg(P9054_HANDLE hPlx, DWORD dwReg, UINT32 dwData)
{
    WD_PCI_CONFIG_DUMP pciCnf;

    BZERO (pciCnf);
    pciCnf.pciSlot = hPlx->pciSlot;
    pciCnf.pBuffer = &dwData;
    pciCnf.dwOffset = dwReg;
    pciCnf.dwBytes = 4;
    pciCnf.fIsRead = FALSE;
    WD_PciConfigDump(hPlx->hWD,&pciCnf);
}

BOOL P9054_DetectCardElements(P9054_HANDLE hPlx)
{
    DWORD i;

    BZERO(hPlx->Int);
    BZERO(hPlx->addrDesc);

    for (i=0; i<hPlx->cardReg.Card.dwItems; i++)
    {
        WD_ITEMS *pItem = &hPlx->cardReg.Card.Item[i];
        DWORD ad_sp;

        switch (pItem->item)
        {
        case ITEM_MEMORY:
            ad_sp = pItem->I.Mem.dwBar;
            hPlx->addrDesc[ad_sp].fIsMemory = TRUE;
            hPlx->addrDesc[ad_sp].dwBytes = pItem->I.Mem.dwBytes;
            hPlx->addrDesc[ad_sp].dwAddr = pItem->I.Mem.dwTransAddr;
            hPlx->addrDesc[ad_sp].dwAddrDirect = pItem->I.Mem.dwUserDirectAddr;
            hPlx->addrDesc[ad_sp].dwMask = hPlx->addrDesc[ad_sp].dwBytes - 1;
            break;
        case ITEM_IO:
            ad_sp = pItem->I.IO.dwBar;
            hPlx->addrDesc[ad_sp].fIsMemory = FALSE;
            hPlx->addrDesc[ad_sp].dwBytes = pItem->I.IO.dwBytes;
            hPlx->addrDesc[ad_sp].dwAddr = pItem->I.IO.dwAddr;
            hPlx->addrDesc[ad_sp].dwMask = hPlx->addrDesc[ad_sp].dwBytes - 1;
            break;
        case ITEM_INTERRUPT:
            if (hPlx->Int.Int.hInterrupt) return FALSE;
            hPlx->Int.Int.hInterrupt = pItem->I.Int.hInterrupt;
            break;
        }
    }

    // Verify that at least one memory space was found
    for (i=0; i<6; i++)
        if (P9054_IsAddrSpaceActive(hPlx, (P9054_ADDR)i)) break;
    if (i==6)
        return FALSE;
    // use address space 0 for accessing local addresses
    hPlx->addrSpace = P9054_ADDR_SPACE0;

    return TRUE;
}

void P9054_Close(P9054_HANDLE hPlx)
{
    // disable interrupts
    if (P9054_IntIsEnabled(hPlx))
        P9054_IntDisable(hPlx);

    // unregister card
    if (hPlx->cardReg.hCard) 
        WD_CardUnregister(hPlx->hWD, &hPlx->cardReg);

    // close WinDriver
    WD_Close(hPlx->hWD);

    free (hPlx);
}

BOOL P9054_IsAddrSpaceActive(P9054_HANDLE hPlx, P9054_ADDR addrSpace)
{
    return hPlx->addrDesc[addrSpace].dwBytes!=0;
}

UINT32 P9054_ReadReg(P9054_HANDLE hPlx, DWORD dwReg)
{
    return P9054_ReadDWord(hPlx, P9054_ADDR_REG, dwReg);
}

void P9054_WriteReg(P9054_HANDLE hPlx, DWORD dwReg, UINT32 dwData)
{
    P9054_WriteDWord(hPlx, P9054_ADDR_REG, dwReg, dwData);
}

BYTE P9054_ReadByte (P9054_HANDLE hPlx, P9054_ADDR addrSpace, DWORD dwOffset)
{
    if (hPlx->addrDesc[addrSpace].fIsMemory)
    {
        DWORD dwAddr = hPlx->addrDesc[addrSpace].dwAddrDirect + dwOffset;
        BYTE *pByte = (BYTE *) dwAddr;
        return *pByte;
    }
    else
    {
        KPTR dwAddr = hPlx->addrDesc[addrSpace].dwAddr + dwOffset;
        WD_TRANSFER trans;
        BZERO(trans);
        trans.cmdTrans = RP_BYTE;
        trans.dwPort = dwAddr;
        WD_Transfer (hPlx->hWD, &trans);
        return trans.Data.Byte;
    }
}

void P9054_WriteByte (P9054_HANDLE hPlx, P9054_ADDR addrSpace, DWORD dwOffset,
    BYTE data)
{
    if (hPlx->addrDesc[addrSpace].fIsMemory)
    {
        DWORD dwAddr = hPlx->addrDesc[addrSpace].dwAddrDirect + dwOffset;
        BYTE *pByte = (BYTE *) dwAddr;
        *pByte = data;
    }
    else
    {
        KPTR dwAddr = hPlx->addrDesc[addrSpace].dwAddr + dwOffset;
        WD_TRANSFER trans;
        BZERO(trans);
        trans.cmdTrans = WP_BYTE;
        trans.dwPort = dwAddr;
        trans.Data.Byte = data;
        WD_Transfer (hPlx->hWD, &trans);
    }
}

WORD P9054_ReadWord (P9054_HANDLE hPlx, P9054_ADDR addrSpace, DWORD dwOffset)
{
    if (hPlx->addrDesc[addrSpace].fIsMemory)
    {
        DWORD dwAddr = hPlx->addrDesc[addrSpace].dwAddrDirect + dwOffset;
        WORD *pWord = (WORD *) dwAddr;
        return dtoh16(*pWord);
    }
    else
    {
        KPTR dwAddr = hPlx->addrDesc[addrSpace].dwAddr + dwOffset;
        WD_TRANSFER trans;
        BZERO(trans);
        trans.cmdTrans = RP_WORD;
        trans.dwPort = dwAddr;
        WD_Transfer (hPlx->hWD, &trans);
        return trans.Data.Word;
    }
}

void P9054_WriteWord (P9054_HANDLE hPlx, P9054_ADDR addrSpace, DWORD dwOffset, WORD data)
{
    if (hPlx->addrDesc[addrSpace].fIsMemory)
    {
        DWORD dwAddr = hPlx->addrDesc[addrSpace].dwAddrDirect + dwOffset;
        WORD *pWord = (WORD *) dwAddr;
        *pWord = dtoh16(data);
    }
    else
    {
        KPTR dwAddr = hPlx->addrDesc[addrSpace].dwAddr + dwOffset;
        WD_TRANSFER trans;
        BZERO(trans);
        trans.cmdTrans = WP_WORD;
        trans.dwPort = dwAddr;
        trans.Data.Word = data;
        WD_Transfer (hPlx->hWD, &trans);
    }
}

UINT32 P9054_ReadDWord (P9054_HANDLE hPlx, P9054_ADDR addrSpace, DWORD dwOffset)
{
    if (hPlx->addrDesc[addrSpace].fIsMemory)
    {
        DWORD dwAddr = hPlx->addrDesc[addrSpace].dwAddrDirect + dwOffset;
        UINT32 *pDword = (UINT32 *) dwAddr;
        return dtoh32(*pDword);
    }
    else
    {
        KPTR dwAddr = hPlx->addrDesc[addrSpace].dwAddr + dwOffset;
        WD_TRANSFER trans;
        BZERO(trans);
        trans.cmdTrans = RP_DWORD;
        trans.dwPort = dwAddr;
        WD_Transfer (hPlx->hWD, &trans);
        return trans.Data.Dword;
    }
}

void P9054_WriteDWord (P9054_HANDLE hPlx, P9054_ADDR addrSpace, DWORD dwOffset, UINT32 data)
{
    if (hPlx->addrDesc[addrSpace].fIsMemory)
    {
        DWORD dwAddr = hPlx->addrDesc[addrSpace].dwAddrDirect + dwOffset;
        UINT32 *pDword = (UINT32 *) dwAddr;
        *pDword = dtoh32(data);
    }
    else
    {
        KPTR dwAddr = hPlx->addrDesc[addrSpace].dwAddr + dwOffset;
        WD_TRANSFER trans;
        BZERO(trans);
        trans.cmdTrans = WP_DWORD;
        trans.dwPort = dwAddr;
        trans.Data.Dword = data;
        WD_Transfer (hPlx->hWD, &trans);
    }
}

void P9054_ReadWriteBlock (P9054_HANDLE hPlx, DWORD dwOffset, PVOID buf, 
    DWORD dwBytes, BOOL fIsRead, P9054_ADDR addrSpace, P9054_MODE mode)
{
    KPTR dwAddr = hPlx->addrDesc[addrSpace].dwAddr + dwOffset;
    WD_TRANSFER trans;

    BZERO(trans);

    if (hPlx->addrDesc[addrSpace].fIsMemory) 
    {
        if (fIsRead) 
        {
            if (mode==P9054_MODE_BYTE) trans.cmdTrans = RM_SBYTE;
            else if (mode==P9054_MODE_WORD) trans.cmdTrans = RM_SWORD;
            else trans.cmdTrans = RM_SDWORD;
        }
        else 
        {
            if (mode==P9054_MODE_BYTE) trans.cmdTrans = WM_SBYTE;
            else if (mode==P9054_MODE_WORD) trans.cmdTrans = WM_SWORD;
            else trans.cmdTrans = WM_SDWORD;
        }
    }
    else 
    {
        if (fIsRead) 
        {
            if (mode==P9054_MODE_BYTE) trans.cmdTrans = RP_SBYTE;
            else if (mode==P9054_MODE_WORD) trans.cmdTrans = RP_SWORD;
            else trans.cmdTrans = RP_SDWORD;
        }
        else 
        {
            if (mode==P9054_MODE_BYTE) trans.cmdTrans = WP_SBYTE;
            else if (mode==P9054_MODE_WORD) trans.cmdTrans = WP_SWORD;
            else trans.cmdTrans = WP_SDWORD;
        }
    }
    trans.dwPort = dwAddr;
    trans.fAutoinc = TRUE;
    trans.dwBytes = dwBytes;
    trans.dwOptions = 0;
    trans.Data.pBuffer = buf;
    WD_Transfer (hPlx->hWD, &trans);
}

void P9054_ReadBlock (P9054_HANDLE hPlx, DWORD dwOffset, PVOID buf, 
    DWORD dwBytes, P9054_ADDR addrSpace, P9054_MODE mode)
{
    P9054_ReadWriteBlock (hPlx, dwOffset, buf, dwBytes, TRUE, addrSpace, mode);
}

void P9054_WriteBlock (P9054_HANDLE hPlx, DWORD dwOffset, PVOID buf, 
    DWORD dwBytes, P9054_ADDR addrSpace, P9054_MODE mode)
{
    P9054_ReadWriteBlock (hPlx, dwOffset, buf, dwBytes, FALSE, addrSpace, mode);
}

void P9054_SetMode (P9054_HANDLE hPlx, DWORD dwLocalAddr)
{
    hPlx->addrDesc[hPlx->addrSpace].dwLocalBase = dwLocalAddr & ~hPlx->addrDesc[hPlx->addrSpace].dwMask;
    hPlx->addrDesc[hPlx->addrSpace].dwLocalBase |= BIT0;
    P9054_WriteReg (hPlx, P9054_LAS0BA, hPlx->addrDesc[hPlx->addrSpace].dwLocalBase);
}

BYTE P9054_ReadByteLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr)
{
    DWORD dwOffset = hPlx->addrDesc[hPlx->addrSpace].dwMask & dwLocalAddr;
    P9054_SetMode (hPlx, dwLocalAddr);
    return P9054_ReadByte(hPlx, hPlx->addrSpace, dwOffset);
}

void P9054_WriteByteLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr, BYTE data)
{
    DWORD dwOffset = hPlx->addrDesc[hPlx->addrSpace].dwMask & dwLocalAddr;
    P9054_SetMode (hPlx, dwLocalAddr);
    P9054_WriteByte(hPlx, hPlx->addrSpace, dwOffset, data);
}

WORD P9054_ReadWordLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr)
{
    DWORD dwOffset = hPlx->addrDesc[hPlx->addrSpace].dwMask & dwLocalAddr;
    P9054_SetMode (hPlx, dwLocalAddr);
    return P9054_ReadWord(hPlx, hPlx->addrSpace, dwOffset);
}

void P9054_WriteWordLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr, WORD data)
{
    DWORD dwOffset = hPlx->addrDesc[hPlx->addrSpace].dwMask & dwLocalAddr;
    P9054_SetMode (hPlx, dwLocalAddr);
    P9054_WriteWord(hPlx, hPlx->addrSpace, dwOffset, data);
}

UINT32 P9054_ReadDWordLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr)
{
    DWORD dwOffset = hPlx->addrDesc[hPlx->addrSpace].dwMask & dwLocalAddr;
    P9054_SetMode (hPlx, dwLocalAddr);
    return P9054_ReadDWord(hPlx, hPlx->addrSpace, dwOffset);
}

void P9054_WriteDWordLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr, UINT32 data)
{
    DWORD dwOffset = hPlx->addrDesc[hPlx->addrSpace].dwMask & dwLocalAddr;
    P9054_SetMode (hPlx, dwLocalAddr);
    P9054_WriteDWord(hPlx, hPlx->addrSpace, dwOffset, data);
}

void P9054_ReadWriteBlockLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr, PVOID buf, 
                    DWORD dwBytes, BOOL fIsRead, P9054_MODE mode)
{
    DWORD dwOffset = hPlx->addrDesc[hPlx->addrSpace].dwMask & dwLocalAddr;
    P9054_SetMode (hPlx, dwLocalAddr);
    P9054_ReadWriteBlock(hPlx, dwOffset, buf, dwBytes, fIsRead, hPlx->addrSpace, mode);
}

void P9054_ReadBlockLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr, PVOID buf, 
                    DWORD dwBytes, P9054_MODE mode)
{
    P9054_ReadWriteBlockLocal (hPlx, dwLocalAddr, buf, dwBytes, TRUE, mode);
}

void P9054_WriteBlockLocal (P9054_HANDLE hPlx, DWORD dwLocalAddr, PVOID buf, 
                     DWORD dwBytes, P9054_MODE mode)
{
    P9054_ReadWriteBlockLocal (hPlx, dwLocalAddr, buf, dwBytes, FALSE, mode);
}

BOOL P9054_IntIsEnabled (P9054_HANDLE hPlx)
{
    if (!hPlx->Int.hThread)
        return FALSE;
    return TRUE;
}

VOID DLLCALLCONV P9054_IntHandler (PVOID pData)
{
    P9054_HANDLE hPlx = (P9054_HANDLE) pData;
    P9054_INT_RESULT intResult;
    intResult.dwCounter = hPlx->Int.Int.dwCounter;
    intResult.dwLost = hPlx->Int.Int.dwLost;
    intResult.fStopped = hPlx->Int.Int.fStopped;
    intResult.dwStatusReg = hPlx->Int.Trans[0].Data.Dword;
    hPlx->Int.funcIntHandler(hPlx, &intResult);  
}

BOOL P9054_IntEnable (P9054_HANDLE hPlx, P9054_INT_HANDLER funcIntHandler)
{
    UINT32 dwIntStatus;
    KPTR dwAddr;
    DWORD dwStatus;

    // check if interrupt is already enabled
    if (hPlx->Int.hThread)
        return FALSE;

    dwIntStatus = P9054_ReadReg (hPlx, P9054_INTCSR);

    BZERO(hPlx->Int.Trans);
    // This is a sample of handling interrupts:
    // Two transfer commands are issued. First the value of the interrupt control/status
    // register is read. Then, a value of ZERO is written.
    // This will cancel interrupts after the first interrupt occurs.
    // When using interrupts, this section will have to change:
    // you must put transfer commands to CANCEL the source of the interrupt, otherwise, the 
    // PC will hang when an interrupt occurs!
    dwAddr = hPlx->addrDesc[AD_PCI_BAR0].dwAddr + P9054_INTCSR;
    hPlx->Int.Trans[0].cmdTrans = hPlx->addrDesc[AD_PCI_BAR0].fIsMemory ? RM_DWORD : RP_DWORD;
    hPlx->Int.Trans[0].dwPort = dwAddr;
    hPlx->Int.Trans[1].cmdTrans = hPlx->addrDesc[AD_PCI_BAR0].fIsMemory ? WM_DWORD : WP_DWORD;
    hPlx->Int.Trans[1].dwPort = dwAddr;
    hPlx->Int.Trans[1].Data.Dword = dwIntStatus & ~(BIT8|BIT10); // put here the data to write to the control register
    hPlx->Int.Int.dwCmds = 2; 
    hPlx->Int.Int.Cmd = hPlx->Int.Trans;
    hPlx->Int.Int.dwOptions |= INTERRUPT_CMD_COPY;

    // this calls WD_IntEnable() and creates an interrupt handler thread
    hPlx->Int.funcIntHandler = funcIntHandler;
    dwStatus = InterruptEnable(&hPlx->Int.hThread, hPlx->hWD, &hPlx->Int.Int, P9054_IntHandler, (PVOID) hPlx);
    if (dwStatus)
    {
        sprintf(P9054_ErrorString, "Enabling interrupts failed with status 0x%lx - %s\n",
            dwStatus, Stat2Str(dwStatus));
        return FALSE;
    }

    // this physically enables interrupts
    P9054_WriteReg (hPlx, P9054_INTCSR, dwIntStatus | (BIT8|BIT10));

    return TRUE;
}

void P9054_IntDisable (P9054_HANDLE hPlx)
{
    DWORD dwIntStatus;

    if (!hPlx->Int.hThread)
        return;

    // this disables interrupts
    dwIntStatus = P9054_ReadReg (hPlx, P9054_INTCSR);
    P9054_WriteReg (hPlx, P9054_INTCSR, dwIntStatus & ~(BIT8|BIT10));

    // this calls WD_IntDisable()
    InterruptDisable(hPlx->Int.hThread);
    hPlx->Int.hThread = NULL;
}

void P9054_DMAReadWriteKernelBlock (P9054_HANDLE hPlx, P9054_DMA_HANDLE hDma, PVOID buf, DWORD dwBytes, BOOL fIsRead)
{
    WD_TRANSFER trans;

    BZERO(trans);

    trans.cmdTrans = fIsRead ? RM_SBYTE : WM_SBYTE;
    trans.dwPort = hDma->dma.pKernelAddr;
    trans.fAutoinc = TRUE;
    trans.dwBytes = dwBytes;
    trans.dwOptions = 0;
    trans.Data.pBuffer = buf;
    WD_Transfer (hPlx->hWD, &trans);
}

P9054_DMA_HANDLE P9054_DMAOpen (P9054_HANDLE hPlx, UINT32 dwLocalAddr, PVOID *buf, DWORD dwDmaOptions, 
    DWORD dwBytes, P9054_MODE mode, P9054_DMA_CHANNEL dmaChannel)
{
    UINT32 dwDMAMODE, dwDMADPR, dwDMALADR;
    UINT32 dwChannelShift = (dmaChannel==P9054_DMA_CHANNEL_0) ? 0 : P9054_DMA_CHANNEL_SHIFT;
    BOOL fAutoinc = FALSE;
    P9054_DMA_HANDLE hDma = NULL;
    DWORD dwStatus, dwPagesNeeded, dwAllocSize;
    BOOL fIsRead;

    if (!(dwDmaOptions & DMA_READ_FROM_DEVICE) && 
        !(dwDmaOptions & DMA_WRITE_TO_DEVICE))
    {
        sprintf(P9054_ErrorString, "9054_DMAOpen error: You must set either "
            "DMA_READ_FROM_DEVICE or DMA_WRITE_TO_DEVICE in the DMA options.\n");
        goto Exit;
    }
    
    if ((dwDmaOptions & DMA_READ_FROM_DEVICE) && 
        (dwDmaOptions & DMA_WRITE_TO_DEVICE))
    {
        sprintf(P9054_ErrorString, "P9054_DMAOpen error: DMA_READ_FROM_DEVICE "
            "and DMA_WRITE_TO_DEVICE were both set.\n");
        goto Exit;
    }
    
    fIsRead = dwDmaOptions & DMA_READ_FROM_DEVICE ? TRUE : FALSE;
    
    // Since the first and last page do not necessarily start (or end) on a 
    // PAGE_SIZE boundary, we need to add 1 more pages for worst case
    dwPagesNeeded = ((dwBytes + GetPageSize() - 1) / GetPageSize()) + 1;

    dwAllocSize = sizeof(P9054_DMA_STRUCT);
    if (WD_DMA_PAGES < dwPagesNeeded)
    {
        dwAllocSize += (sizeof(WD_DMA_PAGE) * (dwPagesNeeded - WD_DMA_PAGES));
        dwDmaOptions |= DMA_LARGE_BUFFER;
    }

    hDma = (P9054_DMA_HANDLE) malloc(dwAllocSize);
    if (hDma==NULL)
    {
        sprintf(P9054_ErrorString, "Failed allocating memory for dma handle!\n");
        goto Exit;
    }
    BZERO (*hDma);
    hDma->dmaChannel = dmaChannel;
    hDma->dma.dwBytes = dwBytes;
    if (!(dwDmaOptions & DMA_KERNEL_BUFFER_ALLOC)) // Scatter/Gather DMA
        hDma->dma.pUserAddr = *buf;
    if (dwDmaOptions & DMA_LARGE_BUFFER)
        hDma->dma.dwPages = dwPagesNeeded;
    hDma->dma.dwOptions = dwDmaOptions;
    hDma->dma.hCard = hPlx->cardReg.hCard;
    dwStatus = WD_DMALock(hPlx->hWD, &hDma->dma);
    if (dwStatus) 
    {
        sprintf(P9054_ErrorString, "Failed locking the buffer. status 0x%lx - %s\n",
            dwStatus, Stat2Str(dwStatus));
        goto Exit;
    }

    // If a contiguous kernel buffer was locked, update buffer:
    if (dwDmaOptions & DMA_KERNEL_BUFFER_ALLOC)
        *buf = hDma->dma.pUserAddr;
    
    // common settings for chain and direct DMA
    dwDMAMODE = 
        (fAutoinc ? 0 : BIT11) 
        | BIT6 // enable Ready input
        | BIT7
        | BIT8 // local burst
        | BIT12
        | (mode==P9054_MODE_BYTE ? 0 : mode==P9054_MODE_WORD ? BIT0 : (BIT1 | BIT0));

    if (hDma->dma.dwPages==1)
    {
        // dma of one page ==> direct dma
        dwDMADPR = BIT0 | (fIsRead ? BIT3 : 0);
        dwDMALADR = dwLocalAddr;

        P9054_WriteReg (hPlx, P9054_DMAMODE + dwChannelShift, dwDMAMODE);
        P9054_WriteReg (hPlx, P9054_DMAPADR + dwChannelShift, (UINT32)hDma->dma.Page[0].pPhysicalAddr);
        P9054_WriteReg (hPlx, P9054_DMALADR + dwChannelShift, dwDMALADR);
        P9054_WriteReg (hPlx, P9054_DMASIZ + dwChannelShift, hDma->dma.Page[0].dwBytes);
        P9054_WriteReg (hPlx, P9054_DMADPR + dwChannelShift, dwDMADPR);
    }
    else
    {
        // dma of more then one page ==> chain dma
        UINT32 dwAlignShift, dwPageNumber, dwMemoryCopied;
        typedef struct {
            UINT32 dwPADR;
            UINT32 dwLADR;
            UINT32 dwSIZ;
            UINT32 dwDPR;
        } DMA_LIST;
        DMA_LIST *pList;
        UINT32 dwStartOfChain;

        // allocate a kernel buffer to hold the chain of DMA descriptors
        // includes extra 0x10 bytes for quadword alignment
        hDma->dmaList.dwBytes = hDma->dma.dwPages * sizeof(DMA_LIST) + 0x10;
        hDma->dmaList.pUserAddr = NULL;
        hDma->dmaList.dwOptions = DMA_KERNEL_BUFFER_ALLOC;
        hDma->dmaList.hCard = hPlx->cardReg.hCard;
        dwStatus = WD_DMALock(hPlx->hWD, &hDma->dmaList);
        if (dwStatus)
        {
            sprintf(P9054_ErrorString, "Failed allocating the chain buffer. status 0x%lx - %s\n",
                dwStatus, Stat2Str(dwStatus));
            goto Exit;
        }

        // verification that bits 0-3 are zero (quadword aligned)
        dwAlignShift = 0x10 - ((UINT32) hDma->dmaList.pUserAddr & 0xf);  
        pList = (DMA_LIST *) ((DWORD) hDma->dmaList.pUserAddr + dwAlignShift);
        dwStartOfChain = (UINT32) hDma->dmaList.Page[0].pPhysicalAddr + dwAlignShift;

        //setting chain of dma pages in the memory
        for (dwPageNumber=0, dwMemoryCopied=0; dwPageNumber<hDma->dma.dwPages; dwPageNumber++)
        {
            pList[dwPageNumber].dwPADR = 
                htod32((UINT32)hDma->dma.Page[dwPageNumber].pPhysicalAddr);
            pList[dwPageNumber].dwLADR = 
                htod32((dwLocalAddr + (fAutoinc ? dwMemoryCopied : 0)));
            pList[dwPageNumber].dwSIZ = 
                htod32((UINT32)hDma->dma.Page[dwPageNumber].dwBytes);
            pList[dwPageNumber].dwDPR = 
                htod32((dwStartOfChain + sizeof(DMA_LIST)*(dwPageNumber+1))
                | BIT0 | (fIsRead ? BIT3 : 0));
            dwMemoryCopied += hDma->dma.Page[dwPageNumber].dwBytes;
        }
        
        pList[dwPageNumber - 1].dwDPR |= htod32(BIT1); // mark end of chain
    
        dwDMAMODE |= BIT9; // chain transfer
        dwDMADPR = dwStartOfChain | BIT0;

        P9054_WriteReg (hPlx, P9054_DMAMODE + dwChannelShift, dwDMAMODE);
        P9054_WriteReg (hPlx, P9054_DMADPR + dwChannelShift, dwDMADPR);
    }

    return hDma;

Exit:
    if (hDma!=NULL)
        P9054_DMAClose(hPlx,hDma);
    return NULL;
}

void P9054_DMAClose (P9054_HANDLE hPlx, P9054_DMA_HANDLE hDma)
{
    if (hDma->dma.hDma)
        WD_DMAUnlock(hPlx->hWD, &hDma->dma);
    if (hDma->dmaList.hDma)
        WD_DMAUnlock(hPlx->hWD, &hDma->dmaList);
    free (hDma);
}

BOOL P9054_DMAIsDone (P9054_HANDLE hPlx, P9054_DMA_HANDLE hDma)
{
    return (P9054_ReadByte(hPlx, P9054_ADDR_REG, P9054_DMACSR + 
        hDma->dmaChannel) & (BYTE)BIT4) == (BYTE)BIT4;
}

void P9054_DMAStart (P9054_HANDLE hPlx, P9054_DMA_HANDLE hDma, BOOL fBlocking)
{
    P9054_WriteByte (hPlx, P9054_ADDR_REG, P9054_DMACSR + hDma->dmaChannel, 
        BIT0 | BIT1);

    //Busy wait for plx to finish transfer
    if (fBlocking) 
        while (!P9054_DMAIsDone(hPlx, hDma))
            ;
}

BOOL P9054_DMAReadWriteBlock (P9054_HANDLE hPlx, DWORD dwLocalAddr, PVOID buf,
    DWORD dwBytes, BOOL fIsRead, P9054_MODE mode, P9054_DMA_CHANNEL dmaChannel)
{
    P9054_DMA_HANDLE hDma;
    DWORD dwOptions = fIsRead ? DMA_READ_FROM_DEVICE : DMA_WRITE_TO_DEVICE;
    if (dwBytes==0) 
        return TRUE;

    // Initiate Scatter/Gather DMA:
    hDma = P9054_DMAOpen(hPlx, dwLocalAddr, &buf, dwOptions, dwBytes, mode, dmaChannel);
    if (!hDma) 
        return FALSE;

    P9054_DMAStart(hPlx, hDma, TRUE);
    P9054_DMAClose(hPlx, hDma);
    return TRUE;
}

BOOL P9054_EEPROMPresent(P9054_HANDLE hPlx)
{
    return (P9054_ReadReg(hPlx, P9054_CNTRL) & BIT28)==BIT28;
}

BOOL P9054_EEPROMValid(P9054_HANDLE hPlx)
{
    WD_PCI_CONFIG_DUMP pciCnf;
    BYTE bVal;

    // check that a blank or programmed serial EEPROM is present
    sprintf(P9054_ErrorString, "serial EEPROM is not present\n");
    if (!P9054_EEPROMPresent(hPlx))
        return FALSE;
        
    // check next capability pointers.
    sprintf(P9054_ErrorString, "capability pointers are incorrect\n");
    BZERO (pciCnf);
    pciCnf.pciSlot = hPlx->pciSlot;
    pciCnf.pBuffer = &bVal;
    pciCnf.fIsRead = TRUE;
    pciCnf.dwBytes = 1;

    pciCnf.dwOffset = PCI_CAP;
    WD_PciConfigDump(hPlx->hWD, &pciCnf);
    if (bVal != (BYTE)P9054_PMCAPID)
        return FALSE;

    pciCnf.dwOffset = P9054_PMNEXT;
    WD_PciConfigDump(hPlx->hWD, &pciCnf);
    if (bVal != (BYTE)P9054_HS_CNTL)
        return FALSE;

    pciCnf.dwOffset = P9054_HS_NEXT;
    WD_PciConfigDump(hPlx->hWD, &pciCnf);
    if (bVal != (BYTE)P9054_VPD_CNTL)
        return FALSE;

    pciCnf.dwOffset = P9054_VPD_NEXT;
    WD_PciConfigDump(hPlx->hWD, &pciCnf);
    if (bVal != 0)
        return FALSE;

    P9054_ErrorString[0] = '\0';
    return TRUE;
}
    
BOOL P9054_EEPROMReadWord(P9054_HANDLE hPlx, DWORD dwOffset, PWORD pwData)
{
    UINT32 dwData;
    DWORD dwAddr;

    if (dwOffset % 2)
    {
        sprintf(P9054_ErrorString, "The offset is not even\n");
        return FALSE;
    }
    dwAddr = dwOffset - (dwOffset % 4);

    if (!P9054_EEPROMReadDWord(hPlx, dwAddr, &dwData))
        return FALSE;

    *pwData = (WORD) (dwData >> ((dwOffset % 4)*8));

    return TRUE;
}

BOOL P9054_EEPROMWriteWord(P9054_HANDLE hPlx, DWORD dwOffset, WORD wData)
{
    UINT32 dwData;
    DWORD dwAddr;

    dwAddr = dwOffset - (dwOffset % 4);

    if (!P9054_EEPROMReadDWord(hPlx, dwAddr, &dwData))
        return FALSE;

    switch (dwOffset % 4)
    {
    case 0:
        dwData = (dwData & 0xffff0000) | wData;
        break;
    case 2:
        dwData = (dwData & 0x0000ffff) | (wData << 16);
        break;
    default:
        sprintf(P9054_ErrorString, "The offset is not even\n");
        return FALSE;
    }

    return P9054_EEPROMWriteDWord(hPlx, dwAddr, dwData);
}

void P9054_Sleep(P9054_HANDLE hPlx, DWORD dwMicroSeconds)
{
    WD_SLEEP sleep;

    BZERO (sleep);
    sleep.dwOptions = SLEEP_NON_BUSY;
    sleep.dwMicroSeconds = dwMicroSeconds;
    WD_Sleep( hPlx->hWD, &sleep);
}

static BYTE P9054_EEPROMRemoveWriteProtection(P9054_HANDLE hPlx, WORD addr)
{
    BYTE old_val;

    old_val = P9054_ReadByte(hPlx, P9054_ADDR_REG, P9054_PROT_AREA);
    
    addr /= 4;
    addr &= 0x7f;

    P9054_WriteByte(hPlx, P9054_ADDR_REG, P9054_PROT_AREA, (BYTE)addr);

    return old_val * 4; //expand from dwords to bytes
}

static void P9054_EEPROMRestoreWriteProtection(P9054_HANDLE hPlx, WORD addr)
{
    P9054_WriteByte(hPlx, P9054_ADDR_REG, P9054_PROT_AREA, (BYTE)addr);
}

static void P9054_EEPROMDataReadWrite(P9054_HANDLE hPlx, BOOL fIsRead, UINT32 *pdwData)
{
    WD_PCI_CONFIG_DUMP pciCnf;

    BZERO (pciCnf);
    pciCnf.pciSlot = hPlx->pciSlot;
    pciCnf.pBuffer = pdwData;
    pciCnf.dwOffset = P9054_VPD_DATA;
    pciCnf.dwBytes = 4;
    pciCnf.fIsRead = fIsRead;
    WD_PciConfigDump(hPlx->hWD,&pciCnf);
}

static void P9054_EEPROMAddrReadWrite(P9054_HANDLE hPlx, BOOL fIsRead, PWORD pwAddr)
{
    WD_PCI_CONFIG_DUMP pciCnf;

    BZERO (pciCnf);
    pciCnf.pciSlot = hPlx->pciSlot;
    pciCnf.pBuffer = pwAddr;
    pciCnf.dwOffset = P9054_VPD_ADDR;
    pciCnf.dwBytes = 2;
    pciCnf.fIsRead = fIsRead;
    WD_PciConfigDump(hPlx->hWD,&pciCnf);
}

BOOL P9054_EEPROMReadDWord(P9054_HANDLE hPlx, DWORD dwOffset, UINT32 *pdwData)
{
    WORD wVal = 0;
    WORD wAddr;
    int i;

    if (dwOffset % 4)
    {
        sprintf(P9054_ErrorString, "The offset is not a multiple of 4\n");
        return FALSE;
    }
    
    // Write a destination serial EEPROM address and flag of operation, value of 0.
    wAddr = (WORD)(dwOffset & ~BIT15);
    P9054_EEPROMAddrReadWrite(hPlx, FALSE, &wAddr);
        
    // Probe a flag of operation until it changes to a 1 to ensure the Read data is available.
    for (i=0; i<10; i++)
    {
        P9054_Sleep(hPlx, P9054_EEPROM_DELAY);
        P9054_EEPROMAddrReadWrite(hPlx, TRUE, &wVal);
        if (wVal & BIT15)
            break;
    }
    
    if (i==10)
    {
        sprintf(P9054_ErrorString, "Acknowledge to EEPROM read was not received\n");
        return FALSE;
    }

    // Read back the requested data from PVPDATA register
    P9054_EEPROMDataReadWrite(hPlx, TRUE, pdwData);

    return TRUE;
}

BOOL P9054_EEPROMWriteDWord(P9054_HANDLE hPlx, DWORD dwOffset, UINT32 dwData)
{
    UINT32 dwReadback;
    WORD wAddr;
    WORD wVal;
    int i;
    BOOL fRet;
    BOOL fReadOk = FALSE;
    BYTE bEnableOffset;

    if (dwOffset % 4)
    {
        sprintf(P9054_ErrorString, "The offset is not a multiple of 4\n");
        return FALSE;
    }

    wAddr = (WORD)dwOffset;
    bEnableOffset = P9054_EEPROMRemoveWriteProtection(hPlx, wAddr);
    P9054_Sleep(hPlx, P9054_EEPROM_DELAY);

    // Write desired data to PVPDATA register
    P9054_EEPROMDataReadWrite(hPlx, FALSE, &dwData);

    // Write a destination serial EEPROM address and flag of operation, value of 1.
    wAddr = (WORD)(wAddr | BIT15);
    P9054_EEPROMAddrReadWrite(hPlx, FALSE, &wAddr);

    // Probe a flag of operation until it changes to a 0 to ensure the write completes.
    for (i=0; i<10; i++)
    {
        P9054_Sleep(hPlx, P9054_EEPROM_DELAY);
        P9054_EEPROMAddrReadWrite(hPlx, TRUE, &wVal);
        if ((wVal & BIT15) == 0)
            break;
    }

    P9054_EEPROMRestoreWriteProtection(hPlx, bEnableOffset); 

    fReadOk = P9054_EEPROMReadDWord(hPlx, dwOffset, &dwReadback);

    if (fReadOk && dwReadback==dwData)
        fRet = TRUE;
    else
    {
        fRet = FALSE;
        if (fReadOk)
            sprintf(P9054_ErrorString, "Write 0x%08x, Read 0x%08x\n",dwData, dwReadback);
        else
            sprintf(P9054_ErrorString, "Error reading EEPROM\n");
    }
    return fRet;
}

// definition of abort bits in commad register (PCI_CR) for PLX 9054
#define P9054_CR_ABORT_BITS (\
    BIT24   /* Detected Master Data Parity Error */ \
    | BIT27 /* Signaled Target Abort */ \
    | BIT28 /* Received Target Abort */ \
    | BIT29 /* Received Master Abort */ \
    | BIT30 /* Signaled System Error */ \
    | BIT31 /* Detected Parity Error on PCI bus */ \
    )

// software reset for PLX 9054 compatible boards
void P9054_reset(P9054_HANDLE hPlx)
{
    BOOL fMUEnabled;
    BOOL fEepromPresent;
    UINT32 temp;
    UINT32 RegCtrl = 0;
    UINT32 RegInterrupt = 0;
    UINT32 RegMailbox0 = 0;
    UINT32 RegMailbox1 = 0;
    UINT32 RegHotSwap = 0;
    UINT32 RegPowerMgmnt = 0;

    // Clear any PCI errors
    RegCtrl = P9054_ReadPCIReg(hPlx, PCI_CR);
    if (RegCtrl & P9054_CR_ABORT_BITS)
    {
        // Write value back to clear aborts
        P9054_WritePCIReg(hPlx, PCI_CR, RegCtrl);
    }

    // Save state of I2O Decode Enable
    temp = P9054_ReadReg(hPlx, P9054_QSR);
    fMUEnabled = temp & BIT0 ? TRUE : FALSE;

    // Determine if an EEPROM is present
    fEepromPresent = P9054_EEPROMPresent(hPlx);

    // Make sure S/W Reset & EEPROM reload bits are clear
    RegCtrl = P9054_ReadReg(hPlx, P9054_CNTRL); 
    RegCtrl &= ~(BIT30 | BIT29);

    // Save some registers if EEPROM present
    if (fEepromPresent)
    {
        // run-time registers
        RegMailbox0 = P9054_ReadReg(hPlx, P9054_MBOX0);
        RegMailbox1 = P9054_ReadReg(hPlx, P9054_MBOX1);

        // PCI configuration registers
        RegInterrupt = P9054_ReadPCIReg(hPlx, PCI_ILR);
        RegHotSwap = P9054_ReadPCIReg(hPlx, P9054_HS_CNTL);
        RegPowerMgmnt = P9054_ReadPCIReg(hPlx, P9054_PMCSR);
    }

    // Issue Software Reset to hold PLX chip in reset
    P9054_WriteReg(hPlx, P9054_CNTRL, RegCtrl | BIT30);

    // Bring chip out of reset
    P9054_Sleep(hPlx, P9054_SOFT_RESET_DELAY);
    P9054_WriteReg(hPlx, P9054_CNTRL, RegCtrl);

    // Issue EEPROM reload in case now programmed
    P9054_WriteReg(hPlx, P9054_CNTRL, RegCtrl | BIT29);

    // Clear EEPROM reload
    P9054_Sleep(hPlx, P9054_SOFT_RESET_DELAY);
    P9054_WriteReg(hPlx, P9054_CNTRL, RegCtrl);

    // Restore I2O Decode Enable state
    if (fMUEnabled)
    {
        temp = P9054_ReadReg(hPlx, P9054_QSR);
        P9054_WriteReg(hPlx, P9054_QSR, temp | BIT0);
    }

    // If EEPROM was present, restore registers
    if (fEepromPresent)
    {
        // run-time registers
        P9054_WriteReg(hPlx, P9054_MBOX0, RegMailbox0);
        P9054_WriteReg(hPlx, P9054_MBOX1, RegMailbox1);

        // PCI configuration registers
        P9054_WritePCIReg(hPlx, PCI_ILR, RegInterrupt);

        // Mask out HS bits that can be cleared
        RegHotSwap &= ~(BIT23 | BIT22 | BIT17);
        P9054_WritePCIReg(hPlx, P9054_HS_CNTL, RegHotSwap);

        // Mask out PM bits that can be cleared
        RegPowerMgmnt &= ~BIT15;
        P9054_WritePCIReg(hPlx, P9054_PMCSR, RegPowerMgmnt);
    }
}

