-------------------------------------------------------------------------------------
-- Entity:	MemoryPortIn
--
-- Project:	Uni1P
--
-- Author:	Antonin Hegar
--
-- Description:
--  Interface between the DSP and BRAMs, where BRAMs are divided into sectors.
--  There are few registers:
--    ADDR_REG 		the DSP sets start address of access to MemoryPort
--  Follows registers are bitmask oriented - bit 0 is sector 0, bit 1 si sector
--  1 etc. To clear a flags be needed to write '1' on matched bit positions
--    DSP_VALID     set by the DSP when in the sector are data
--                  clear by the FPGA's application when data are processed
--
-- Revision:
--	$Log: MemoryPortIn.vhd,v $
--	Revision 1.2  2006/04/12 05:30:15  Tony
--	Bug glitch appears when FpgaSectorValid fall edge - fixed.
--	
--	Revision 1.1  2006/01/24 20:58:36  Tony
--	Initial version.
--	Implemented with Xilinx ISE 7.1.04i
--	
--
-------------------------------------------------------------------------------------

-------------------------------------------------------------------------------------
-- LIBRARY
-------------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.DX64Header.all;
use work.MathLib.all;

-------------------------------------------------------------------------------------
-- ENTITY
-------------------------------------------------------------------------------------
entity MemoryPortIn is
	generic (
        SECTOR_FIRST_C  : natural := 0;                         -- number of first sector in the port
        SECTOR_LAST_C   : natural := 3;                         -- number of last sector in the port
		
    		DATA_WIDTH_C 	: natural := DSP_DATA_WIDTH_C;  						-- data port width of BRAM
        ADDR_WIDTH_C    : natural := 11                         -- addr port width
	);
	port (
		-----------------------------------------------------------------------------
		-- global signals
		AReset			: in std_logic;
		Clock			: in std_logic;
		
        Debug           : out unsigned(31 downto 0);
        
        -----------------------------------------------------------------------------
		-- DSP interface
		DspBus			: inout DspBus_T;
		
        -----------------------------------------------------------------------------
		-- Chaining with other MemoryPorts
		ChainNext 		: in MemoryPortChain_T;
		ChainPrev 		: out MemoryPortChain_T;
		
        -----------------------------------------------------------------------------
		-- BRAM interface
		BramAddr 		: in unsigned(ADDR_WIDTH_C-1 downto 0);
		BramDataOut		: out unsigned(DATA_WIDTH_C-1 downto 0);
		
        DspSectorValid  : out unsigned(SECTOR_LAST_C downto SECTOR_FIRST_C);
        DspSectorClear  : in  unsigned(SECTOR_LAST_C downto SECTOR_FIRST_C)
        
	);
end entity MemoryPortIn;


-------------------------------------------------------------------------------------
-- ENTITY
-------------------------------------------------------------------------------------
architecture RTL of MemoryPortIn is

    ---------------------------------------------------------------------------
    -- Components declaration
    ---------------------------------------------------------------------------
    attribute box_type : string;
    component RAMB16_S18_S18
        generic (
            SIM_COLLISION_CHECK : string := "ALL";
            SRVAL_A : bit_vector := X"00000";
            SRVAL_B : bit_vector := X"00000";
            WRITE_MODE_A : string := "WRITE_FIRST";
            WRITE_MODE_B : string := "WRITE_FIRST"
        );
        port (
            DOA   : out std_logic_vector(15 downto 0);
            DOB   : out std_logic_vector(15 downto 0);
            DOPA  : out std_logic_vector(1 downto 0);
            DOPB  : out std_logic_vector(1 downto 0);
            ADDRA : in  std_logic_vector(9 downto 0);
            ADDRB : in  std_logic_vector(9 downto 0);
            CLKA  : in  std_ulogic;
            CLKB  : in  std_ulogic;
            DIA   : in  std_logic_vector(15 downto 0);
            DIB   : in  std_logic_vector(15 downto 0);
            DIPA  : in  std_logic_vector(1 downto 0);
            DIPB  : in  std_logic_vector(1 downto 0);
            ENA   : in  std_ulogic;
            ENB   : in  std_ulogic;
            SSRA  : in  std_ulogic;
            SSRB  : in  std_ulogic;
            WEA   : in  std_ulogic;
            WEB   : in  std_ulogic);
    end component;
    attribute box_type of RAMB16_S18_S18 : component is "black_box";
    
    
	---------------------------------------------------------------------------------
	-- Type definition and constants
	---------------------------------------------------------------------------------
    -- number of sectors in the port
    constant SECTORS_COUNT_C    : natural := SECTOR_LAST_C-SECTOR_FIRST_C+1;
    -- boundary MSB addresses of sectors in this port
    constant MSB_ADDR_LOW_C     : natural := SECTOR_FIRST_C * MEMPORT_SECTOR_SIZE_C * DSP_DATA_WIDTH_C / BRAM_SIZE_BITS_C;
    constant MSB_ADDR_HIGH_C    : natural := SECTOR_LAST_C  * MEMPORT_SECTOR_SIZE_C * DSP_DATA_WIDTH_C / BRAM_SIZE_BITS_C;
	
    
	---------------------------------------------------------------------------------
	-- Signal definition
	---------------------------------------------------------------------------------
	signal data_out			    : unsigned(DSP_DATA_WIDTH_C-1 downto 0);
	
    ---------------------------------------------------------------------------
    -- BRAM signals
    -- number of BRAMs allocated in the port
    constant BRAM_COUNT_C       : natural := MEMPORT_SECTOR_SIZE_C * SECTORS_COUNT_C * DSP_DATA_WIDTH_C / BRAM_SIZE_BITS_C;
    constant BRAMA_ADDR_WIDTH_C : natural := log2(BRAM_SIZE_BITS_C/DATA_WIDTH_C - 1);
    constant BRAMB_ADDR_WIDTH_C : natural := log2(BRAM_SIZE_BITS_C/DSP_DATA_WIDTH_C - 1);
	
	
    type BramaDataoutArr_T is array (0 to BRAM_COUNT_C-1) of std_logic_vector(DATA_WIDTH_C-1 downto 0);
    type BrambDataoutArr_T is array (0 to BRAM_COUNT_C-1) of std_logic_vector(DSP_DATA_WIDTH_C-1 downto 0);    
    signal brama_datain         : std_logic_vector(DATA_WIDTH_C-1 downto 0);
	signal brama_dataout        : BramaDataoutArr_T;
    signal brama_addr           : std_logic_vector(BRAMA_ADDR_WIDTH_C-1 downto 0);
    signal brama_wr_select      : unsigned(BRAM_COUNT_C-1 downto 0);
    signal brama_rd_nselect     : unsigned(BRAM_COUNT_C-1 downto 0);
    signal brama_dataout_or     : std_logic_vector(DATA_WIDTH_C-1 downto 0) := (others => '0');
	
    signal bramb_datain         : std_logic_vector(DSP_DATA_WIDTH_C-1 downto 0);
	signal bramb_dataout        : BrambDataoutArr_T;
    signal bramb_addr           : std_logic_vector(BRAMB_ADDR_WIDTH_C-1 downto 0);
    signal bramb_wr_select      : unsigned(BRAM_COUNT_C+MSB_ADDR_LOW_C-1 downto MSB_ADDR_LOW_C);
    signal bramb_rd_nselect     : unsigned(BRAM_COUNT_C+MSB_ADDR_LOW_C-1 downto MSB_ADDR_LOW_C);
    signal bramb_dataout_or     : std_logic_vector(DSP_DATA_WIDTH_C-1 downto 0) := (others => '0');
    signal bramb_dataout_valid  : std_logic;
    
    ---------------------------------------------------------------------------
    -- Internal registers
    signal sector_ready_reg     : unsigned(SECTOR_LAST_C downto SECTOR_FIRST_C) := (others => '0');
    signal sector_ovf_reg       : unsigned(SECTOR_LAST_C downto SECTOR_FIRST_C) := (others => '0');
    signal sector_mask_reg    	: unsigned(SECTOR_LAST_C downto SECTOR_FIRST_C) := (others => '0');    
	
	signal sector_ready_read    : unsigned(MEMPORT_MAX_SECTORS_C-1 downto 0) := (others => '0');
    signal sector_ovf_read      : unsigned(MEMPORT_MAX_SECTORS_C-1 downto 0) := (others => '0');
	
    signal ready        	   : std_logic;
    signal ovf      	       : std_logic;
    
--    constant FPGA_ADDR_WIDTH_C  : natural := log2(MEMPORT_SECTOR_SIZE_C * SECTORS_COUNT_C * DSP_DATA_WIDTH_C / DATA_WIDTH_C - 1);
    constant SECTORS_ADDR_WIDTH_C:natural := BRAMB_ADDR_WIDTH_C + log2(MEMPORT_MAX_SECTORS_C - 1);
    signal dsp_addr             : unsigned(SECTORS_ADDR_WIDTH_C-1 downto 0) := (others => '0');
    
    ---------------------------------------------------------------------------
    -- adresses signals
    signal sel_ready_low    	: std_logic;
    signal sel_ready_high  	 	: std_logic;    
    signal sel_ovf_low     		: std_logic;
    signal sel_ovf_high    		: std_logic;
    signal sel_mask_low    		: std_logic;
    signal sel_mask_high   		: std_logic;
    signal sel_data             : std_logic;
    signal sel_addr             : std_logic;
    
begin
	
    ---------------------------------------------------------------------------
    -- GLOBAL ASSERTIONS
    ---------------------------------------------------------------------------
    assert DATA_WIDTH_C = 16
        report "MemoryPort: unsupported value of DATA_WIDTH_C parameter!"
        severity FAILURE;
    
    assert MEMPORT_SECTOR_SIZE_C * SECTORS_COUNT_C * DSP_DATA_WIDTH_C REM BRAM_SIZE_BITS_C = 0
        report "MemoryPort: Allocated memory has to be multiple by BRAM_SIZE_BITS_C!"
        severity FAILURE;
	
    assert MEMPORT_SECTOR_SIZE_C * SECTORS_COUNT_C * DSP_DATA_WIDTH_C / DATA_WIDTH_C = 2**ADDR_WIDTH_C
        report "MemoryPort: Size of allocated memory has to be same as BramAddr space (2**ADDR_WIDTH_C)!"
        severity FAILURE;
	
    
    ---------------------------------------------------------------------------
    -- BRAM instance
    ---------------------------------------------------------------------------
    Memory_G: for i in 0 to BRAM_COUNT_C-1 generate
        
		
        -----------------------------------------------------------------------
        -- DATA_WIDTH_C = 16
        -----------------------------------------------------------------------
        MemoryWidth16_G: if DATA_WIDTH_C = 16 generate
		
            RAMB16_S18_S18_G : RAMB16_S18_S18
                generic map (
                    SRVAL_A  => "000000000000000000",
                    SRVAL_B  => "000000000000000000"
                )
                port map (
                    DOA   => brama_dataout(i),
                    DOB   => bramb_dataout(i),
                    DOPA  => OPEN,
                    DOPB  => OPEN,
                    ADDRA => brama_addr,
                    ADDRB => bramb_addr,
                    CLKA  => Clock,
                    CLKB  => Clock,
                    DIA   => brama_datain,
                    DIB   => bramb_datain,
                    DIPA  => "00",
                    DIPB  => "00",
                    ENA   => '1',
                    ENB   => '1',
                    SSRA  => brama_rd_nselect(i),
                    SSRB  => bramb_rd_nselect(i+MSB_ADDR_LOW_C),
                    WEA   => brama_wr_select(i),
                    WEB   => bramb_wr_select(i+MSB_ADDR_LOW_C)
                );
        end generate MemoryWidth16_G;
		
    end generate Memory_G;
    ---------------------------------------------------------------------------
    -- BRAM connections
    ---------------------------------------------------------------------------
    bramb_addr      <= std_logic_vector( dsp_addr(BRAMB_ADDR_WIDTH_C-1 downto 0) );
    bramb_datain    <= std_logic_vector( DspBus.DataIn );
    
    BramDataout_P: process (brama_dataout, bramb_dataout)
        variable brama_t    : std_logic_vector(DATA_WIDTH_C-1 downto 0);
        variable bramb_t    : std_logic_vector(DSP_DATA_WIDTH_C-1 downto 0);
    begin
        brama_t := (others => '0');
        bramb_t := (others => '0');
        for i in 0 to BRAM_COUNT_C-1 loop
            bramb_t     := bramb_t OR bramb_dataout(i);
            brama_t     := brama_t OR brama_dataout(i);
        end loop;
        bramb_dataout_or <= bramb_t;
        brama_dataout_or <= brama_t;
    end process BramDataout_P;
	
    ---------------------------------------------------------------------------
    -- Sector contents valid data
    ---------------------------------------------------------------------------
    SectorValid_P: process (AReset, Clock)
    begin
        if AReset = '1' then
            sector_ready_reg   	<= (others => '1');
			sector_ovf_reg 		<= (others => '0');
        elsif Clock'event AND Clock = '1' then
            for i in SECTOR_FIRST_C to SECTOR_LAST_C loop
				-- flag is in the first halfword
				if i < DSP_DATA_WIDTH_C then
					if DspBus.Write = '1' AND sel_ready_low = '1' then
						sector_ready_reg(i) <= (NOT DspBus.DataIn(i)) AND sector_ready_reg(i);
					else
						if DspSectorClear(i) = '1' then
							sector_ready_reg(i) <= '1';
						end if;
					end if;
				-- flag is in the second halfword
				else
					if DspBus.Write = '1' AND sel_ready_high = '1' then
						sector_ready_reg(i) <= (NOT DspBus.DataIn(i-DSP_DATA_WIDTH_C)) AND sector_ready_reg(i);
					else
						if DspSectorClear(i) = '1' then
							sector_ready_reg(i) <= '1';
						end if;
					end if;
				end if;
            end loop;                   -- i
        end if;
    end process SectorValid_P;
	
    ---------------------------------------------------------------------------
    -- Sector mask register 
    ---------------------------------------------------------------------------
    SectorMaskReg_P: process (AReset, Clock)
    begin
        if AReset = '1' then
            sector_mask_reg <= (others => '0');
        elsif Clock'event AND Clock = '1' then
            for i in SECTOR_FIRST_C to SECTOR_LAST_C loop
				---------------------------------------------------------------
				-- Mask bits
                -- flag is in the first halfword
                if i < DSP_DATA_WIDTH_C then
                    if DspBus.Write = '1' AND sel_mask_low = '1' then
                        sector_mask_reg(i) <= DspBus.DataIn(i);
                    end if;
                -- flag is in the second halfword
                else
                    if DspBus.Write = '1' AND sel_mask_high = '1' then
                        sector_mask_reg(i) <= DspBus.DataIn(i-DSP_DATA_WIDTH_C);
                    end if;
                end if;
				
			end loop;                   -- i
            
        end if;
	end process SectorMaskReg_P;
	
	
    ---------------------------------------------------------------------------
    -- The DSP side address
    ---------------------------------------------------------------------------
    DspAddr_P: process (AReset, Clock)
    begin
        if AReset = '1' then
            dsp_addr <= (others => '0');
        elsif Clock'event AND Clock = '1' then
            if DspBus.Write = '1' AND sel_addr = '1' then
                dsp_addr <= DspBus.DataIn(SECTORS_ADDR_WIDTH_C-1 downto 0);
            elsif (DspBus.Write = '1' OR DspBus.Read = '1') AND sel_data = '1' then
                dsp_addr <= dsp_addr + 1;
            end if;
        end if;
    end process DspAddr_P;
	
    ---------------------------------------------------------------------------
    -- Translate DSP address to local port BRAM
    --  bramb_wr_select
    --  bramb_rd_nselect
    ---------------------------------------------------------------------------
    BrambSelect_P: process (DspBus, dsp_addr, sel_data)
        variable sel     : natural;
    begin
        sel := to_integer( dsp_addr(SECTORS_ADDR_WIDTH_C-1 downto BRAMB_ADDR_WIDTH_C) );
        bramb_wr_select     <= (others => '0');
        bramb_rd_nselect    <= (others => '1');        
        if MSB_ADDR_LOW_C <= sel AND sel <= MSB_ADDR_HIGH_C then
            bramb_wr_select(sel) <= DspBus.Write AND sel_data;
            bramb_rd_nselect(sel) <= NOT (DspBus.Read AND sel_data);
        end if;
    end process BrambSelect_P;
	
    ---------------------------------------------------------------------------
    -- Translate FPGA address to local port BRAM
    --  brama_wr_select
    --  brama_rd_nselect
    ---------------------------------------------------------------------------
    BramaSelect_P: process (BramAddr)
        variable sel        : natural range 0 to BRAM_COUNT_C-1;
    begin
        if BRAM_COUNT_C > 1 then
            sel := to_integer( BramAddr(ADDR_WIDTH_C-1 downto BRAMA_ADDR_WIDTH_C) );
            brama_wr_select         <= (others => '0');
            brama_rd_nselect        <= (others => '1');        
            brama_wr_select(sel)    <= '0';
            brama_rd_nselect(sel)   <= '0';
        else
            brama_wr_select(0)      <= '0';
            brama_rd_nselect(0)     <= '0';
        end if;
        brama_addr              <= std_logic_vector( BramAddr(BRAMA_ADDR_WIDTH_C-1 downto 0) );
    end process BramaSelect_P;
    brama_datain        <= (others => '0');
    BramDataOut         <= unsigned( brama_dataout_or );
	
    ---------------------------------------------------------------------------
    -- Format registers to read by the DSP
	-- sector_ready 	- '1' indicates ready sector for the DSP
	--  					if the sector is input then it is full - the FPGA is written data
	--  					if the sector is output then it is empty - the FPGA is read the data
    ---------------------------------------------------------------------------
    ReadFormat_P: process (sector_ready_reg, sector_ovf_reg)
    begin
		sector_ready_read  	<= (others => '0');
        sector_ovf_read   	<= (others => '0');
        sector_ready_read(SECTOR_LAST_C downto SECTOR_FIRST_C) 	<= sector_ready_reg;
        sector_ovf_read(SECTOR_LAST_C downto SECTOR_FIRST_C)  	<= sector_ovf_reg;
    end process ReadFormat_P;
    
	---------------------------------------------------------------------------------
	-- DataOut to the DSP
	DataOut_P: process (AReset, Clock)
	begin
        if AReset = '1' then
            data_out <= (others => '0');
		elsif Clock'event AND Clock = '1' then
			-- read sector ready flags - low bits
			if DspBus.Read = '1' and  sel_ready_low = '1' then
                data_out <= sector_ready_read(DSP_DATA_WIDTH_C-1 downto 0);
			-- read sector ready flags - high bits
			elsif DspBus.Read = '1' and  sel_ready_high = '1' then
                data_out <= sector_ready_read(MEMPORT_MAX_SECTORS_C-1 downto DSP_DATA_WIDTH_C);
			-- read sector ovf flags - low bits
			elsif DspBus.Read = '1' and  sel_ovf_low = '1' then
                data_out <= sector_ovf_read(DSP_DATA_WIDTH_C-1 downto 0);
			-- read sector ovf flags - high bits
			elsif DspBus.Read = '1' and  sel_ovf_high = '1' then
                data_out <= sector_ovf_read(MEMPORT_MAX_SECTORS_C-1 downto DSP_DATA_WIDTH_C);
			-- read overrun flags
			else
				data_out <= (others => '0');
			end if;
		end if;
	end process DataOut_P;
	
    ---------------------------------------------------------------------------
    -- Valid data on output of BRAM (the DSP side) - one clock latency after read signal
    ---------------------------------------------------------------------------
    DataoutValid_P: process (AReset, Clock)
	begin
		if AReset = '1' then
			bramb_dataout_valid <= '0';
		elsif Clock'event AND Clock = '1' then
            if DspBus.Read = '1' AND sel_data = '1' then
                bramb_dataout_valid <= '1';
            else
                bramb_dataout_valid <= '0';
            end if;
		end if;
	end process DataoutValid_P;
	
	DspBus.DataOut <= unsigned(bramb_dataout_or) when bramb_dataout_valid = '1' else data_out;
	
    ---------------------------------------------------------------------------
    -- Address decoder
    ---------------------------------------------------------------------------
    sel_ready_low  <= '1' when DspBus.Addr = MEMPORT_READY_LOW_ADDR_C else '0';
    sel_ready_high <= '1' when DspBus.Addr = MEMPORT_READY_HIGH_ADDR_C else '0';    
    sel_mask_low   <= '1' when DspBus.Addr = MEMPORT_MASK_LOW_ADDR_C else '0';
    sel_mask_high  <= '1' when DspBus.Addr = MEMPORT_MASK_HIGH_ADDR_C else '0';    
    sel_ovf_low    <= '1' when DspBus.Addr = MEMPORT_OVF_LOW_ADDR_C else '0';
    sel_ovf_high   <= '1' when DspBus.Addr = MEMPORT_OVF_HIGH_ADDR_C else '0';
	sel_data	   <= '1' when DspBus.Addr = MEMPORT_DATA_LOW_ADDR_C OR DspBus.Addr = MEMPORT_DATA_HIGH_ADDR_C  else '0';
    sel_addr       <= '1' when DspBus.Addr = MEMPORT_ADDR_LOW_ADDR_C else '0';
	
    ---------------------------------------------------------------------------
    -- GLOBAL ASSIGNMENT
    ---------------------------------------------------------------------------
    DspSectorValid      <= NOT sector_ready_reg;
	
    ---------------------------------------------------------------------------
    -- Sum of all valid and ovf flags in this port
    ---------------------------------------------------------------------------
    ValidOvfSum_P: process (sector_ready_reg, sector_mask_reg, sector_ovf_reg)
        variable ready_t, ovf_t     : std_logic;
    begin
        ready_t   := '0';
        ovf_t     := '0';
        for i in SECTOR_FIRST_C to SECTOR_LAST_C loop
            ready_t   := ready_t OR (sector_ready_reg(i) AND sector_mask_reg(i));
            ovf_t     := ovf_t OR sector_ovf_reg(i);
        end loop;  -- i
        ready  <= ready_t;
        ovf    <= ovf_t;
    end process ValidOvfSum_P;
    
    ---------------------------------------------------------------------------
    -- If memory port is first, then generate registers on chain
    ---------------------------------------------------------------------------
    FirstChain_G: if SECTOR_FIRST_C = 0 generate
		
        ChainRegOut_P: process (AReset, Clock)
        begin
            if AReset = '1' then
                ChainPrev.SectorReady   <= '0';
                ChainPrev.SectorOvf     <= '0';
            elsif Clock'event AND Clock = '1' then
                ChainPrev.SectorReady   <= ChainNext.SectorReady  or ready;
                ChainPrev.SectorOvf   	<= ChainNext.SectorOvf    or ovf;
            end if;
        end process ChainRegOut_P;
        
    end generate FirstChain_G;
	
    ---------------------------------------------------------------------------
    -- If memory port is second or more, then generate combinational on chain
    ---------------------------------------------------------------------------
    Chain_G: if SECTOR_FIRST_C > 0 generate
		
        ChainPrev.SectorReady   <= ChainNext.SectorReady  or ready;
        ChainPrev.SectorOvf     <= ChainNext.SectorOvf    or ovf;
        
    end generate Chain_G;
    
    Debug(10 downto 0) <= dsp_addr(10 downto 0);
	
end RTL;
