
-- Module Name:    wrapper - Behavioral 
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use work.DX64Header.all;
use work.MathLib.all;

entity wrapper_8b is
	 generic (
		 C_LATENCY : integer := 2;
		 C_RD_SEC : integer := 4;
		 C_WR_SEC : integer := 8;
		 MEM_IN_ADDR_WD : integer := 11;
		 MEM_OUT_ADDR_WD : integer := 12;
		 MEM_IN_DATA_WD : integer := 16;
		 MEM_OUT_DATA_WD : integer := 16
	 );
   port( 
		  clk : in  std_logic;
      rst : in  std_logic;

      mem_in_addr : out unsigned(MEM_IN_ADDR_WD-1 downto 0);      
      valid_sector_in : in unsigned(C_RD_SEC-1 downto 0); 
      clear_sector_in : out unsigned(C_RD_SEC-1 downto 0);
      hilo : out std_logic; 

      data_out_valid : out std_logic;
      mem_out_addr : out unsigned(MEM_OUT_ADDR_WD-1 downto 0);
      valid_sector_out : out unsigned(C_WR_SEC-1 downto 0);
      clear_sector_out : in unsigned(C_WR_SEC-1 downto 0);
      --sobel
      ena : out std_logic
	);
end wrapper_8b;

architecture Behavioral of wrapper_8b is
-- CONSTANT DEFINITION
----------------------------------------------------------
constant SEC_IN_NUM_WD : natural := log2(C_RD_SEC);
constant SEC_OUT_NUM_WD : natural := log2(C_WR_SEC);

-- SIGNAL DEFINITION
----------------------------------------------------------

--delayed valid signal
signal data_valid_sh_reg : std_logic_vector(C_LATENCY-1 downto 0);
--adress is one input sector
signal internal_sec_in_addr : unsigned(MEM_IN_ADDR_WD-1 downto 0);
--adress is one otput sector
signal internal_sec_out_addr : unsigned(MEM_OUT_ADDR_WD-1 downto 0);
-- outpu data valid
signal internal_data_out_valid : std_logic;
-- sector address offset in MemoryPortIn
signal mem_addr_in_offset : unsigned(MEM_IN_ADDR_WD-1 downto 0);
signal mem_addr_out_offset : unsigned(MEM_OUT_ADDR_WD-1 downto 0);
-- input data valid
signal data_in_valid : std_logic;
-- clerar actual sector
signal sec_in_clear : std_logic;
-- number of actual sector
signal sec_in_num : std_logic_vector(SEC_IN_NUM_WD-1 downto 0);
signal sec_out_num : std_logic_vector(SEC_OUT_NUM_WD-1 downto 0);

--output sector valid reg
signal internal_valid_sector_out : unsigned(C_WR_SEC-1 downto 0);
signal shadow_valid_sector_out : unsigned(C_WR_SEC-1 downto 0);

-- Sobel Enable
signal sobel_ena : std_logic;

-- Byte selection
signal high_byte : std_logic;

begin


-- BEHAVIOARAL --
-------------------------------------------------------------------------

-- OUTPUT DATA VALID
--------------------------------------------------------------------------------
  -- Process: output_valid_delay 
  output_valid_delay: process (clk, rst) is
  begin
    if rst = '1' then 
      data_valid_sh_reg <= (others => '0');
    elsif clk = '1' and clk'event then
      for i in 0 to C_LATENCY-2 loop
        data_valid_sh_reg(i+1) <= data_valid_sh_reg(i);
      end loop;
      data_valid_sh_reg(0) <= data_in_valid;    
    end if;
  end process output_valid_delay;

-- MEMORY (SECTORS) ADDRESSING
--------------------------------------------------------------------------------


  -- Process: low_high_byte 
  low_high_byte: process (clk, rst) is
  begin
    if rst = '1' then 
      high_byte <= '0';
    elsif clk = '1' and clk'event then
      if (data_in_valid = '1') then
        high_byte <= not high_byte;
      else
        high_byte <= '0';
      end if;    
    end if;
  end process low_high_byte;
  


  -- Process: input_memory_address 
  input_memory_address: process (clk, rst) is
  begin
    if rst = '1' then 
      internal_sec_in_addr <= (others => '0');
      sobel_ena <= '0';
    elsif clk = '1' and clk'event then
      if (data_in_valid = '1') then
        if (internal_sec_in_addr = ((2* MEMPORT_SECTOR_SIZE_C / (MEM_IN_DATA_WD / 8))-1) and high_byte = '1' ) then
          internal_sec_in_addr <= (others => '0');
        elsif (high_byte = '1') then
          internal_sec_in_addr <= internal_sec_in_addr + 1;
        else
          internal_sec_in_addr <= internal_sec_in_addr;
        end if;
        sobel_ena <= '1';        
      else 
        internal_sec_in_addr <= (others => '0');
        sobel_ena <= '0';      
      end if;
    end if;
  end process input_memory_address;
  
  -- Process: output_memory_address 
  output_memory_address: process (clk, rst) is
  begin
    if rst = '1' then 
      internal_sec_out_addr <= (others => '0');
    elsif clk = '1' and clk'event then
      if(internal_data_out_valid = '1') then
        if (internal_sec_out_addr = ((2 * MEMPORT_SECTOR_SIZE_C / (MEM_OUT_DATA_WD / 8))-1)) then
          internal_sec_out_addr <= (others => '0');
        else
          internal_sec_out_addr <= internal_sec_out_addr + 1;
        end if;
      else 
        internal_sec_out_addr <= (others => '0');
      end if;
    end if;
  end process output_memory_address;
  
  
-- INPUT SECTORS VALIDATION / CLEAR
--------------------------------------------------------------------------------
  --If any input sector contain valid data
  -- Process: data_in_valid_process 
  data_in_valid_process: process (valid_sector_in) is
  variable rdy_val : std_logic;
  begin
    rdy_val := '0';
    for i in 0 to C_RD_SEC-1 loop
      rdy_val := rdy_val or valid_sector_in(i);
    end loop;
    data_in_valid <= rdy_val;
  end process data_in_valid_process;
  
  --convert first valid sector to memory space addres offset
  -- Process: sector_in_address_decoder 
  sector_in_address_decoder: process (valid_sector_in) is
  begin
    mem_addr_in_offset <= (others => '0');
    sec_in_num <= (others => '0');
    for i in 0 to C_RD_SEC-1 loop
      exit when valid_sector_in(i)='1';
      --one sector contains 512 B and memory interface has 'MEM_IN_DATA_WD' data width interface
      --offset is address where starts first valid sector
      mem_addr_in_offset <= conv_unsigned((i+1) * (2 * MEMPORT_SECTOR_SIZE_C / (MEM_IN_DATA_WD / 8)), MEM_IN_ADDR_WD);
      sec_in_num <= std_logic_vector(conv_unsigned((i+1) , SEC_IN_NUM_WD));
    end loop;
  end process sector_in_address_decoder;


  -- Process: sec_in_clr_process 
  sec_in_clr_process: process (sec_in_num, sec_in_clear) is
  begin
      for i in 0 to C_RD_SEC-1 loop
        if (i = sec_in_num) then 
          clear_sector_in(i) <= sec_in_clear;
        else
          clear_sector_in(i) <= '0';
        end if;
      end loop;
  end process sec_in_clr_process;
  
  
-- OUTPUT SECTORS VALIDATION / CLEAR
--------------------------------------------------------------------------------

  -- Process: sec_out_valid_process 
  sec_out_valid_process: process (clk, rst) is
  begin
    if rst = '1' then 
      shadow_valid_sector_out <= (others => '0');
      internal_valid_sector_out <= (others => '0');
    elsif clk = '1' and clk'event then
      for i in 0 to C_WR_SEC-1 loop
        if internal_data_out_valid = '1' then
          if i = sec_out_num then
            shadow_valid_sector_out(i) <= '1';
            if (internal_sec_out_addr = ((2 * MEMPORT_SECTOR_SIZE_C / (MEM_OUT_DATA_WD / 8))-1)) then
              internal_valid_sector_out(i) <= '1';
            end if;
          else
             shadow_valid_sector_out(i) <= shadow_valid_sector_out(i) and not clear_sector_out(i);
             internal_valid_sector_out(i) <= internal_valid_sector_out(i) and not clear_sector_out(i);
          end if;
        else
          shadow_valid_sector_out(i) <= '0';
          internal_valid_sector_out(i) <= ((shadow_valid_sector_out(i) or internal_valid_sector_out(i)) and not clear_sector_out(i));
        end if;        
      end loop;
    end if;
  end process sec_out_valid_process;
  
  --convert first empty sector to memory space addres offset
  -- Process: sector_out_address_decoder 
  sector_out_address_decoder: process (internal_valid_sector_out) is
  begin
    mem_addr_out_offset <= (others => '0');
    sec_out_num <= (others => '0');
    for i in C_WR_SEC-1 downto 0 loop
      -- no address update when full sector is valid
      exit when internal_valid_sector_out(i)='1';
      -- address update
      mem_addr_out_offset <= conv_unsigned((i) * (2 * MEMPORT_SECTOR_SIZE_C / (MEM_OUT_DATA_WD / 8)), MEM_OUT_ADDR_WD);
      sec_out_num <= std_logic_vector(conv_unsigned((i) , SEC_OUT_NUM_WD));
      -- stop address update when sector is updated
      exit when shadow_valid_sector_out(i)='1';
    end loop;
  end process sector_out_address_decoder;
    
  
  
-- GLOBAR ASSERTIONS
--------------------------------------------------------------------------------  
  --output valid
  internal_data_out_valid <= data_valid_sh_reg(C_LATENCY-1);
  data_out_valid <= internal_data_out_valid;

  
  --clear input sector
  sec_in_clear <= '1' when (internal_sec_in_addr = ((2 * MEMPORT_SECTOR_SIZE_C / (MEM_IN_DATA_WD / 8))-1) and high_byte='1') else '0';

  --memory adresses
  mem_out_addr <= internal_sec_out_addr + mem_addr_out_offset;
  mem_in_addr <= internal_sec_in_addr + mem_addr_in_offset;
  
  -- output sector valid reg 
  valid_sector_out <= internal_valid_sector_out;
  
--   data_out <= unsigned(sobel_data_out);
--   sobel_data_in <=std_logic_vector(data_in);
  
  
  ena <= sobel_ena;
  
  hilo <= high_byte;
end Behavioral;
